/*
    contours-cor.c - Edouard Thiel - 29/05/2001

    Compilation :
        cc bsutil.c contours-cor.c -o contours-cor \
            `~/helium/helium-cfg --cflags --libs` -lm

    Excution :    contours-cor image.pgm
*/

#include <helium.h>
#include <math.h>
#include "bsutil.h"


He_node *princ, *canvas, *panel, *text1, *text2, *tog1, *tog2;
int c_gris, c_rouge, c_vert, c_bleu, c_jaune, c_cyan, c_magenta;
BsMap *bm_orig = NULL, *bm_res1 = NULL, *bm_res2 = NULL;

enum {  AFFI_ORIGINAL, AFFI_POINT_CONTOUR, AFFI_SUIVI_CONTOUR,
        AFFI_APPROX_POLYG, AFFI_RELAXATION, AFFI_TRACE_POLYG,
        AFFI_VORONOI_DISCRET };
int etat_affi = AFFI_ORIGINAL;

enum {  DIST_4, DIST_8, DIST_34, DIST_5711 };
int etat_dist = DIST_34;

#define POLYG_VECMAX 20000
int polyg_x[POLYG_VECMAX],
    polyg_y[POLYG_VECMAX],
    polyg_d[POLYG_VECMAX],
    polyg_n = 0;

#define CHA_MAX 100
int cha_x[CHA_MAX], cha_y[CHA_MAX], cha_w[CHA_MAX], cha_n = 0;


/*--------------------------- A F F I C H A G E ------------------------------*/

#define GROS_POINT 3
#define MIL_POINT (GROS_POINT/2)
#define MAX_WIDTH 1000
#define MAX_HEIGHT 700

void init_couleurs ()
{
    c_gris    = HeAllocRgb (150, 150, 150, he_white);
    c_rouge   = HeAllocRgb (255, 0, 0, he_black);
    c_vert    = HeAllocRgb (0, 255, 0, he_black);
    c_bleu    = HeAllocRgb (0, 0, 255, he_black);
    c_jaune   = HeAllocRgb (255, 255, 0, he_black);
    c_cyan    = HeAllocRgb (0, 255, 255, he_black);
    c_magenta = HeAllocRgb (255, 0, 255, he_black);
}


/*
 * Dessin d'un gros point
*/

void dessin_point (Window win, int x, int y, int coul)
{
    XSetForeground (he_display, he_gc, coul);
    XFillRectangle (he_display, win, he_gc,
        x * GROS_POINT, y * GROS_POINT, GROS_POINT, GROS_POINT);
}


/*
 * Dessin d'un trait fin dont les extrmits sont centres
 * sur les gros points.
*/

void dessin_trait (Window win, int x1, int y1, int x2, int y2, int coul)
{
    XSetForeground (he_display, he_gc, coul);
    XDrawLine (he_display, win, he_gc,
        x1 * GROS_POINT + MIL_POINT, y1 * GROS_POINT + MIL_POINT,
        x2 * GROS_POINT + MIL_POINT, y2 * GROS_POINT + MIL_POINT);
}


/* Affichage point par point en noir/gris :
 *   0 : noir,  >0 : gris
 * (pas aussi rapide que par une XImage, mais suffisant ici)
*/
void dessin_image_binaire (Window win, BsMap *bm)
{
    int x, y;

    for (y = 0; y < bm->ysiz; y++)
    for (x = 0; x < bm->xsiz; x++)
	dessin_point (win, x, y,
            bm->tab[y][x] == 0 ? he_black : c_gris);
}


/* Affichage point par point en noir/gris/couleurs (modulo 6):
 *   0 noir, 1 gris, 2 bleu, 3 jaune, 4 cyan, 5 magenta, 6 rouge, 7 vert, ...
*/
void dessin_image_couleur (Window win, BsMap *bm)
{
    int x, y, v, v6, c;

    for (y = 0; y < bm->ysiz; y++)
    for (x = 0; x < bm->xsiz; x++) {
        v = bm->tab[y][x];
        v6 = v % 6;
        c = v  == 0 ? he_black  :
            v  == 1 ? c_gris    :
            v6 == 0 ? c_rouge   :
            v6 == 1 ? c_vert    :
            v6 == 2 ? c_bleu    :
            v6 == 3 ? c_jaune   :
            v6 == 4 ? c_cyan    :
            v6 == 5 ? c_magenta :
            he_white;

	dessin_point (win, x, y, c);
    }
}


/*
 * Dessine la polygonalisation avec des traits fins en alternance vert/rouge.
 * On relie les sommets tels que polyg_d[] > 0 ; on ne relie pas les sommets
 * tels que polyg_d[] sont != , pour sparer les diffrents contours.
*/

void dessin_traits_polyg (Window win)
{
    int i, k = -1, c = c_vert;

    for (i = 0; i < polyg_n; i++) {
        if (polyg_d[i] == 0) continue;
        if (polyg_d[k] == polyg_d[i])
             dessin_trait (win, polyg_x[k], polyg_y[k],
	         polyg_x[i], polyg_y[i], c);
        k = i;
        c = c == c_vert ? c_rouge : c_vert;
    }

}


/*-------------------------- A L G O R I T H M E S ---------------------------*/

/* Directions de Freeman, en                   3 2 1
   tournant dans le sens inverse               4 p 0
   des aiguilles d'une montre.                 5 6 7
*/
int nx[] = { 1, 1, 0, -1, -1, -1, 0, 1 },
    ny[] = { 0, -1, -1, -1, 0, 1, 1, 1 };


/*
 * Met le fond  0, la forme  1, le contour  2
 * on considre la connexit 8 pour la forme et 4 pour le fond.
 * Les images ont un bord extrieur dj initialis  0.
*/

void calcul_points_contour (BsMap *d, BsMap *r)
{
    int x, y, i, trouve;

    for (y = 0; y < d->ysiz; y++)
    for (x = 0; x < d->xsiz; x++)
    {
        if (d->tab[y][x] == 0)
            r->tab[y][x] = 0;
        else {
            trouve = FALSE;
            for (i = 0; i < 8; i+=2)
                if (d->tab[y+ny[i]][x+nx[i]] == 0) {
		    trouve = TRUE; break;
                }
            r->tab[y][x] = trouve ? 2 : 1;
        }
    }
}


/* Suivi d'un contour  partir de xP,yP et de la direction fP ;
 * on mmorise le contour dans polyg_x,y et on tiqute  nc
 * dans polyg_d et l'image r
*/

void Suivi1C (BsMap *r, int xP, int yP, int fP, int nc)
{
    int x, y, i, k, kFin, trouve;

    /* Init : on marque le premier point */
    x = xP; y = yP;
    r->tab[y][x] = nc;
    polyg_x[polyg_n] = x;
    polyg_y[polyg_n] = y;
    polyg_d[polyg_n] = nc;
    polyg_n++;

    /* Recherche de la direction kFin dans laquelle on reviendra sur le point
     * final xP,yP : c'est la direction du premier point  partir de fP en
     * tournant dans le sens des aiguilles d'une montre.
    */
    trouve = FALSE;
    for (i = 0, k = fP; i < 8; i++, k = (k+7)%8)
	if (r->tab[y+ny[k]][x+nx[k]] != 0) { trouve = TRUE; break;}

    if (!trouve) return;  /* Point isol : on limine le cas */
    k = kFin = (k+4)%8;   /* direction d'o l'on vient */

    /* Suivi du contour en tournant dans le sens inverse des aiguilles
     * d'une montre ; k est la direction courante.
    */
    do {
        /* Recherche du point suivant ; il y en a obligatoirement un
           puisque le point actuel n'est pas isol */
        for (i = 0, k = (k+5)%8; i < 8; i++, k = (k+1)%8)
            if (r->tab[y+ny[k]][x+nx[k]] != 0) break;

        /* On marque le nouveau point */
        x = x+nx[k]; y = y+ny[k];
        r->tab[y][x] = nc;
        polyg_x[polyg_n] = x;
        polyg_y[polyg_n] = y;
        polyg_d[polyg_n] = nc;
        polyg_n++;

      /* Condition d'arrt : on arrive sur xP,yP avec la direction kFin */
    } while (!(x == xP && y == yP && k == kFin));
}


/*
 * Met le fond  0, la forme  1, le 1er contour  2, le suivant  3, etc
 * Stocke les points du contour dans polyg_x,y[]
 * Stocke le numero du contour dans polyg_d[] (2, 3, ...)
 *
 * Les images ont un bord extrieur dj initialis  0.
 *
 * Restriction : pour que le contour d'un trou soit dtect, il doit possder
 * au moins un point qui n'appartient pas au contour extrieur.
*/

void calcul_suivi_contour (BsMap *d, BsMap *r)
{
    int x, y, nc = 2, i,
        ni[] = { 4, 2, 0, 6};  /* ordre de parcours des directions de Freeman */
                               /* permettant de dtecter d'abord l'extrieur. */
    /* init */
    polyg_n = 0;
    for (y = 0; y < d->ysiz; y++)
    for (x = 0; x < d->xsiz; x++)
        r->tab[y][x] = d->tab[y][x] == 0 ? 0 : 1;

    for (y = 0; y < d->ysiz; y++)
    for (x = 0; x < d->xsiz; x++)
        if (r->tab[y][x] == 1)
        {
            for (i = 0; i < 4; i++)
                if (r->tab[y+ny[ni[i]]][x+nx[ni[i]]] == 0)
                    { Suivi1C (r, x, y, ni[i], nc++); break; }
        }
}


/*
 * Renvoie la distance euclidienne du point C  la droite (AB)
*/

double dist_point_droite (int xC, int yC, int xA, int yA, int xB, int yB)
{
    double x1 = xB-xA, y1 = yB-yA, x2 = xC-xA, y2 = yC-yA,
           d = x1*x1 + y1*y1;
    return d == 0 ? sqrt (x2*x2 + y2*y2)
    		  : fabs (x1*y2 - x2*y1) / sqrt (d);
}


/*
 * Renvoie l'indice i, i compris entre a et b, du point le plus loign de la
 * droite passant par les points d'indice a et b.
*/

int point_plus_distant (int a, int b)
{
    int i, k = a;
    double v, m = -1;

    for (i = a; i <= b; i++) {
        v = dist_point_droite (polyg_x[i], polyg_y[i],
            polyg_x[a], polyg_y[a], polyg_x[b], polyg_y[b]);
        if (m < v) { m = v; k = i; }
    }
    return k;
}


/*
 * Dcoupage polygonal rcursif de a..b avec le seuil s
*/

void decoupe_rec (int a, int b, double s)
{
    int i, k;

    k = point_plus_distant (a, b);

    if (dist_point_droite (polyg_x[k], polyg_y[k],
            polyg_x[a], polyg_y[a], polyg_x[b], polyg_y[b]) > s)
    {
        decoupe_rec (a, k, s);
        decoupe_rec (k, b, s);
    } else {
        /* Marque tous les points de a+1..b-1  0 (non-sommet) */
        for (i = a+1; i < b; i++)
            polyg_d[i] = 0;
    }
}


/*
 * Calcule l'approximation polygonale,  partir de polyg_x,y[]
 * Cela consiste  mettre les points non retenus comme sommets  0
 * dans polyg_d[]
*/

void calcul_approx_polyg (double s)
{
    int i, a = 0, b = 0, nc = 0;

    for (i = 0; i < polyg_n; i++) {
        if (polyg_d[i] == 0) continue;

        if (polyg_d[i] == nc) b = i;
        else {
            if (nc > 0) decoupe_rec (a, b, s);
            nc = polyg_d[i]; a = b = i;
        }
    }
    /* dernier contour */
    if (nc > 0) decoupe_rec (a, b, s);
}


/*
 * Fait la relaxation sur l'approximation polygonale :
 * cela consiste  mettre  0 certains sommets jugs inutiles dans polyg_d[]
*/

void calcul_relax_polyg (double s)
{
    int i, a = -1, b = -1, c = -1, nc = 0;

    for (i = 0; i < polyg_n; i++) {
        if (polyg_d[i] == 0) continue;

        if (polyg_d[i] == nc) {
            /* on dcale a,b,c -> b,c,i */
            a = b; b = c; c = i;

            /* test si on n'a pas 3 points a,b,c */
            if (a == -1) continue;

            /* test si on peut supprimer b */
	    if (dist_point_droite (polyg_x[b], polyg_y[b],
		polyg_x[a], polyg_y[a], polyg_x[c], polyg_y[c]) <= s)
            {
                polyg_d[b] = 0; b = a;
            }

        } else {
            /* nouveau contour */
            nc = polyg_d[i]; a = -1; b = -1; c = i;
        }
    }
}


/*
 * Colorie les points mmoriss dans polyg_x,y[] en changeant de
 * couleur chaque fois que polyg_d[] != 0
*/

void calcul_trace_polyg (BsMap *r)
{
    int i, k = 2;

    for (i = 0; i < polyg_n; i++) {
        r->tab[polyg_y[i]][polyg_x[i]] = k;
        if (polyg_d[i] != 0) k++;
    }
}


/*
 * Affecte les poids du masque de chanfrein
*/

void fixe_masque_chanfrein ()
{
    cha_n = 0;

    switch (etat_dist) {
        case DIST_4 :
            cha_x[cha_n] =  1; cha_y[cha_n] =  0; cha_w[cha_n++] =  1;
            cha_x[cha_n] =  0; cha_y[cha_n] =  1; cha_w[cha_n++] =  1;
            break;
        case DIST_8 :
            cha_x[cha_n] =  1; cha_y[cha_n] =  0; cha_w[cha_n++] =  1;
            cha_x[cha_n] =  0; cha_y[cha_n] =  1; cha_w[cha_n++] =  1;
            cha_x[cha_n] =  1; cha_y[cha_n] =  1; cha_w[cha_n++] =  1;
            cha_x[cha_n] = -1; cha_y[cha_n] =  1; cha_w[cha_n++] =  1;
            break;
        case DIST_34 :
            cha_x[cha_n] =  1; cha_y[cha_n] =  0; cha_w[cha_n++] =  3;
            cha_x[cha_n] =  0; cha_y[cha_n] =  1; cha_w[cha_n++] =  3;
            cha_x[cha_n] =  1; cha_y[cha_n] =  1; cha_w[cha_n++] =  4;
            cha_x[cha_n] = -1; cha_y[cha_n] =  1; cha_w[cha_n++] =  4;
            break;
        case DIST_5711 :
            cha_x[cha_n] =  1; cha_y[cha_n] =  0; cha_w[cha_n++] =  5;
            cha_x[cha_n] =  0; cha_y[cha_n] =  1; cha_w[cha_n++] =  5;
            cha_x[cha_n] =  1; cha_y[cha_n] =  1; cha_w[cha_n++] =  7;
            cha_x[cha_n] = -1; cha_y[cha_n] =  1; cha_w[cha_n++] =  7;
            cha_x[cha_n] = -2; cha_y[cha_n] =  1; cha_w[cha_n++] = 11;
            cha_x[cha_n] =  2; cha_y[cha_n] =  1; cha_w[cha_n++] = 11;
            cha_x[cha_n] = -1; cha_y[cha_n] =  2; cha_w[cha_n++] = 11;
            cha_x[cha_n] =  1; cha_y[cha_n] =  2; cha_w[cha_n++] = 11;
            break;
    }
}


/*
 * Calcule le Diagramme de Voronoi gnralis discret avec
 * une distance de chanfrein sur l'image de l'approximation polygonale.
 *
 * Les images ont un bord extrieur suffisant, dj initialis  0.
 *
 * vor = image des contours (fond  0, intrieur = 1, contours > 1)
 *       dans laquelle on fait le Voronoi Discret
 * idt = image dans laquelle on fait la transformation de distances
*/

void calcul_voronoi_discret (BsMap *vor, BsMap *idt)
{
    int i, k, x, y, mini, v;

    /* Init de idt : intrieur de la forme  1, bord et extrieur  0.
    */
    for (y = 0; y < idt->ysiz; y++)
    for (x = 0; x < idt->xsiz; x++)
	idt->tab[y][x] = (vor->tab[y][x] != 1) ? 0 : 1;

    /* Premier balayage */
    for (y = 0; y < idt->ysiz; y++)
    for (x = 0; x < idt->xsiz; x++)
        if (idt->tab[y][x] != 0)
        {
            mini = 30000; k = 0;
            for (i = 0; i < cha_n; i++)
            {
                v = idt->tab[y - cha_y[i]][x - cha_x[i]] + cha_w[i];
                if (v < mini) { mini = v; k = i; }
            }
            idt->tab[y][x] = mini;
            vor->tab[y][x] = vor->tab[y - cha_y[k]][x - cha_x[k]];
        }

    /* Second balayage */
    for (y = idt->ysiz-1; y >= 0; y--)
    for (x = idt->xsiz-1; x >= 0; x--)
        if (idt->tab[y][x] != 0)
        {
            mini = idt->tab[y][x]; k = -1;
            for (i = 0; i < cha_n; i++)
            {
                v = idt->tab[y + cha_y[i]][x + cha_x[i]] + cha_w[i];
                if (v < mini) { mini = v; k = i; }
            }
            if (k != -1) 
            {
                idt->tab[y][x] = mini;
                vor->tab[y][x] = vor->tab[y + cha_y[k]][x + cha_x[k]];
            }
        }
}


/*
 * On recalcule tout  chaque changement d'tat : a vite de devoir
 * cliquer sur les boutons radio dans un certain ordre.
*/

void applique_filtres ()
{
    double seuil = atof(HeGetTextValue(text2));

    /* Scurit */
    if (bm_orig == NULL) return;
    InitExterBsMap (bm_orig, 0);

    switch (etat_affi) {

        case AFFI_POINT_CONTOUR : {
            calcul_points_contour (bm_orig, bm_res1);
        } break;

        case AFFI_SUIVI_CONTOUR : {
            calcul_suivi_contour (bm_orig, bm_res1);
        } break;

        case AFFI_APPROX_POLYG : {
            calcul_suivi_contour (bm_orig, bm_res1);
            calcul_approx_polyg (seuil);
        } break;

        case AFFI_RELAXATION : {
            calcul_suivi_contour (bm_orig, bm_res1);
            calcul_approx_polyg (seuil);
            calcul_relax_polyg (seuil);
       } break;

        case AFFI_TRACE_POLYG : {
            calcul_suivi_contour (bm_orig, bm_res1);
            calcul_approx_polyg (seuil);
            calcul_relax_polyg (seuil);
            calcul_trace_polyg (bm_res1);
        } break;

        case AFFI_VORONOI_DISCRET : {
            calcul_suivi_contour (bm_orig, bm_res1);
            calcul_approx_polyg (seuil);
            calcul_relax_polyg (seuil);
            calcul_trace_polyg (bm_res1);
	    fixe_masque_chanfrein ();
            calcul_voronoi_discret (bm_res1, bm_res2);
        } break;

    }

    /* MAJ affichage */
    HePostRepaint (canvas);
}


/*--------------------------- C A L L B A C K S ------------------------------*/

void tog1_proc (He_node *hn)
{
    char *nom = HeGetToggleLabel (hn);

    if (HeGetToggleValue (hn) == FALSE) return;

    if (strcmp (nom, "Original") == 0)
        etat_affi = AFFI_ORIGINAL;
    else if (strcmp (nom, "Points contour") == 0)
        etat_affi = AFFI_POINT_CONTOUR;
    else if (strcmp (nom, "Suivi contour") == 0)
        etat_affi = AFFI_SUIVI_CONTOUR;
    else if (strcmp (nom, "Approx polygonale") == 0)
        etat_affi = AFFI_APPROX_POLYG;
    else if (strcmp (nom, "Relaxation") == 0)
        etat_affi = AFFI_RELAXATION;
    else if (strcmp (nom, "Trac") == 0)
        etat_affi = AFFI_TRACE_POLYG;
    else if (strcmp (nom, "Voronoi discret") == 0)
        etat_affi = AFFI_VORONOI_DISCRET;

    applique_filtres ();
}


void tog2_proc (He_node *hn)
{
    char *nom = HeGetToggleLabel (hn);

    if (HeGetToggleValue (hn) == FALSE) return;

    if (strcmp (nom, "d4") == 0)
        etat_dist = DIST_4;
    else if (strcmp (nom, "d8") == 0)
        etat_dist = DIST_8;
    else if (strcmp (nom, "d3,4") == 0)
        etat_dist = DIST_34;
    else if (strcmp (nom, "d5,7,11") == 0)
        etat_dist = DIST_5711;

    applique_filtres ();
}


void canvas_repaint (He_node *hn, Window win)
{
    if (bm_orig == NULL) return;

    switch (etat_affi) {

        case AFFI_ORIGINAL : {
            dessin_image_binaire (win, bm_orig);
        } break;

        case AFFI_POINT_CONTOUR : {
            dessin_image_couleur (win, bm_res1);
        } break;

        case AFFI_SUIVI_CONTOUR : {
            dessin_image_couleur (win, bm_res1);
        } break;

        case AFFI_APPROX_POLYG : {
            dessin_image_binaire (win, bm_orig);
	    dessin_traits_polyg (win);
        } break;

        case AFFI_RELAXATION : {
            dessin_image_binaire (win, bm_orig);
	    dessin_traits_polyg (win);
        } break;

        case AFFI_TRACE_POLYG : {
            dessin_image_couleur (win, bm_res1);
        } break;

        case AFFI_VORONOI_DISCRET : {
            dessin_image_couleur (win, bm_res1);
        } break;

    }
}


void quit_proc (He_node *hn)
{
    HeQuit(0);
}


void lire_proc (He_node *hn)
{
    char *nomf = HeGetTextValue (text1);
    int k;

    /* Destruction anciennes donnes */
    FreeBsMap (bm_orig); bm_orig = NULL;
    FreeBsMap (bm_res1); bm_res1 = NULL;
    FreeBsMap (bm_res2); bm_res2 = NULL;

    /* Lecture */
    bm_orig = ReadBsMapFromPGM (nomf, 2);

    /* Test russite */
    if (bm_orig == NULL) {
        HeSimpleDialog (
            HE_DIALOG_BELL,
            HE_DIALOG_TITLE,   "Attention",
            HE_DIALOG_MESSAGE, "Erreur de lecture du fichier",
            HE_DIALOG_QUOTED,  nomf,
            HE_DIALOG_BUTTON,  "OK",
            0);
	return;
    }

    /* Cration de bm_res1 et bm_res2  la taille de bm_orig */
    bm_res1 = NewBsMap (bm_orig->ysiz, bm_orig->xsiz, 2);
    bm_res2 = NewBsMap (bm_orig->ysiz, bm_orig->xsiz, 2);
    if (bm_res1 == NULL || bm_res2 == NULL) HeQuit(1);

    /* Taille canvas et fentre */
    k = bm_orig->xsiz * GROS_POINT; if (k > MAX_WIDTH) k = MAX_WIDTH;
    HeSetWidth (canvas, k);
    k = bm_orig->ysiz * GROS_POINT; if (k > MAX_HEIGHT) k = MAX_HEIGHT;
    HeSetHeight (canvas, k);
    HeSetX (canvas, 0);
    HeFit (princ);
    HeSetX (canvas, (HeGetWidth(princ) - HeGetWidth(canvas)-2)/2);

    applique_filtres ();
}


void text2_proc (He_node *hn)
{
    applique_filtres ();
}


/*--------------------------------- M A I N ----------------------------------*/

int main (int argc, char **argv)
{
    int code_sortie;

    /* Init Helium, ouverture display X11, dcodage arguments */
    HeInit (&argc, &argv);

    /* Les arguments qui restent n'ont pas t reconnus par Helium */
    if (argc != 2){
        printf ("USAGE : %s image.pgm\n", argv[0]);
        HeQuit(1);
    }

    /* Cration widgets */
    princ = HeCreateFrame ();
    HeSetFrameLabel (princ, "Contours (cor)");

    panel = HeCreatePanel (princ);

    HeCreateMessageP (panel, "Image binaire :", TRUE);
    text1 = HeCreateText (panel);
    HeSetTextVisibleLen (text1, 40);
    HeSetTextValue (text1, argv[1]);
    HeSetTextCompletion (text1, TRUE);
    HeSetTextNotifyProc (text1, lire_proc);
    HeCreateButtonP (panel, "Lire", lire_proc, NULL);
    HeCreateButtonP (panel, "Quit", quit_proc, NULL);

    HeSetPanelLayout (panel, HE_LINE_FEED);
    HeCreateMessageP (panel, "Afficher :", TRUE);
    tog1 = HeCreateToggleRadioP (panel, "Original", tog1_proc, NULL);
    HeCreateToggleRadioP (panel, "Points contour", tog1_proc, tog1);
    HeCreateToggleRadioP (panel, "Suivi contour", tog1_proc, tog1);

    HeSetPanelLayout (panel, HE_LINE_FEED);
    HeCreateToggleRadioP (panel, "Approx polygonale", tog1_proc, tog1);
    HeCreateToggleRadioP (panel, "Relaxation", tog1_proc, tog1);
    HeCreateToggleRadioP (panel, "Trac", tog1_proc, tog1);
    HeCreateToggleRadioP (panel, "Voronoi discret", tog1_proc, tog1);

    HeSetPanelLayout (panel, HE_LINE_FEED);
    HeCreateMessageP (panel, "Seuil :", TRUE);
    text2 = HeCreateText (panel);
    HeSetTextVisibleLen (text2, 8);
    HeSetTextValue (text2, "1.0");
    HeSetTextNotifyProc (text2, text2_proc);
    HeCreateMessageP (panel, " Distance :", TRUE);
    tog2 = HeCreateToggleRadioP (panel, "d4", tog2_proc, NULL);
    HeCreateToggleRadioP (panel, "d8", tog2_proc, tog2);
    HeCreateToggleRadioP (panel, "d3,4", tog2_proc, tog2);
    HeCreateToggleRadioP (panel, "d5,7,11", tog2_proc, tog2);

    HeFit (panel);

    canvas = HeCreateCanvas (princ);
    HeSetCanvasRepaintProc (canvas, canvas_repaint);
    HeSetY (canvas, HeGetHeight(panel) + 5);

    HeFit (princ);

    /* Lecture et affichage */
    init_couleurs ();
    lire_proc (NULL);

    /* Boucle d'vnements */
    code_sortie = HeMainLoop (princ);

    /* Le display X11 est maintenant ferm ; on libre la mmoire */
    FreeBsMap (bm_orig);
    FreeBsMap (bm_res1);
    FreeBsMap (bm_res2);

    return code_sortie;
}
